import pytest
from unittest import mock

from eotdl.eotdl.auth.auth import Auth
from eotdl.src.errors.auth import LoginError, AuthTimeOut
from eotdl.eotdl.auth.is_logged import IsLogged
from eotdl.eotdl.auth.logout import Logout


def test_auth():
    repo = mock.Mock()
    api_repo = mock.Mock()
    auth = Auth(repo, api_repo)
    inputs = Auth.Inputs()
    api_repo.login.return_value = mock.Mock(
        status_code=200,
        json=mock.Mock(return_value={"code": 123, "login_url": "123123"}),
    )
    token_data = {"id_token": "lkajshdflkjahsldkfjhalskjdfhlkasjdhf"}
    api_repo.token.return_value = mock.Mock(
        status_code=200, json=mock.Mock(return_value=token_data)
    )
    repo.save_creds.return_value = "creds_path"
    repo.decode_token.return_value = {
        "uid": "test",
        "name": "test name",
        "email": "test email",
        "picture": "test picture",
    }
    outputs = auth(inputs)
    api_repo.login.assert_called_once()
    api_repo.token.assert_called_once_with(123)
    repo.save_creds.assert_called_once_with(token_data)
    repo.decode_token.assert_called_once_with(token_data)
    assert outputs.user["uid"] == "test"
    assert outputs.user["name"] == "test name"
    assert outputs.user["email"] == "test email"
    assert outputs.user["picture"] == "test picture"


def test_auth_fails_if_not_logged_in():
    repo = mock.Mock()
    api_repo = mock.Mock()
    auth = Auth(repo, api_repo)
    inputs = Auth.Inputs()
    api_repo.login.return_value = mock.Mock(status_code=400)
    with pytest.raises(LoginError):
        auth(inputs)
    api_repo.login.assert_called_once()


def test_auth_fails_if_auth_timeout():
    repo = mock.Mock()
    api_repo = mock.Mock()
    auth = Auth(
        repo, api_repo, max_t=0.1, interval=0.01
    )  # will try to login 10 times, then timeout
    inputs = Auth.Inputs()
    api_repo.login.return_value = mock.Mock(
        status_code=200,
        json=mock.Mock(return_value={"code": 123, "login_url": "123123"}),
    )
    api_repo.token.return_value = mock.Mock(status_code=400)
    with pytest.raises(AuthTimeOut):
        auth(inputs)


def test_is_logged_in():
    repo = mock.Mock()
    repo.load_creds.return_value = {"uid": 123}
    is_logged = IsLogged(repo)
    inputs = IsLogged.Inputs()
    outputs = is_logged(inputs)
    repo.load_creds.assert_called_once()
    assert outputs.user == {"uid": 123}


def test_is_not_logged_in():
    repo = mock.Mock()
    repo.load_creds.return_value = None
    is_logged = IsLogged(repo)
    inputs = IsLogged.Inputs()
    outputs = is_logged(inputs)
    repo.load_creds.assert_called_once()
    assert outputs.user is None


def test_logout():
    repo = mock.Mock()
    api_repo = mock.Mock()
    logout = Logout(repo, api_repo)
    inputs = Logout.Inputs()
    api_repo.logout_url.return_value = "ñklasjdf"
    outputs = logout(inputs)
    api_repo.logout_url.assert_called_once()
    repo.logout.assert_called_once()
    assert outputs.logout_url == "ñklasjdf"



"""
Parameters to access Airbus data
"""


class AirbusURL:
    """
    AIRBUS URLs
    """
    PRICES = "https://data.api.oneatlas.airbus.com/api/v1/prices"
    ORDERS = "https://data.api.oneatlas.airbus.com/api/v1/orders"
    SEARCH = "https://search.foundation.api.oneatlas.airbus.com/api/v2/opensearch?constellation=SPOT"
    ALL_ORDERS_STATUS = "https://data.api.oneatlas.airbus.com/api/v1/orders"
    ACCOUNT = "https://data.api.oneatlas.airbus.com/api/v1/me"
    ROLES = "https://data.api.oneatlas.airbus.com/api/v1/me/services"


# Types are defined at: https://www.geoapi-airbusds.com/api-catalog-v2/oad-living-library/tutorials/#order-an-individual-product


class AirbusProductType:
    """
    Airbus product types
    """
    MULTISPECTRAL = "multiSpectral"


class AirbusRadiometricProcessing:
    """
    Airbus radiometric processing
    """
    REFLECTANCE = "REFLECTANCE"


class AirbusImageFormat:
    """
    Airbus image format
    """
    GEOTIFF = "image/geotiff"
    JP2 = "image/jp2"



import pytest
import os
import json
from eotdl.tools import get_item_metadata, remove_raster_metadata


# Create a fixture for a temporary directory and sample files
@pytest.fixture
def temp_dir(tmpdir):
    raster_file = tmpdir.join("sample.tif")
    metadata_file = tmpdir.join("metadata.json")
    same_name_metadata_file = tmpdir.join("sample.json")
    
    # Create a sample metadata content
    metadata_content = {"sample_key": "sample_value"}

    with open(metadata_file, 'w') as f:
        json.dump(metadata_content, f)
        
    with open(same_name_metadata_file, 'w') as f:
        json.dump(metadata_content, f)
    
    return str(tmpdir), str(raster_file)


def test_get_item_metadata(temp_dir):
    dir_path, raster_path = temp_dir
    metadata = get_item_metadata(raster_path)
    
    assert metadata is not None
    assert metadata["sample_key"] == "sample_value"

    # Additional tests can be to delete the 'metadata.json' and ensure the code picks up 'sample.json'
    os.remove(os.path.join(dir_path, 'metadata.json'))
    metadata = get_item_metadata(raster_path)
    assert metadata is not None

    # Further, you can remove 'sample.json' and ensure the function returns None
    os.remove(os.path.join(dir_path, 'sample.json'))
    assert get_item_metadata(raster_path) is None


def test_remove_raster_metadata(temp_dir):
    dir_path, _ = temp_dir
    remove_raster_metadata(dir_path)
    remove_raster_metadata(dir_path, 'sample.json')
    
    # Assert that the metadata files are removed
    assert not os.path.exists(os.path.join(dir_path, 'metadata.json'))
    assert not os.path.exists(os.path.join(dir_path, 'sample.json'))



import os
import pystac

from eotdl.curation.stac.extensions import ScaneoLabeler, ImageNameLabeler

from .fixtures import tmp_stac_catalog, tmp_stac_catalog_labels, EXAMPLE_DATA_DIR

scaneo_labeler = ScaneoLabeler()
image_labeler = ImageNameLabeler()


def test_scaneo_labeler(tmp_stac_catalog):
    from pystac.extensions.label import LabelExtension

    labels_extra_properties = {"label_methods": ["manual"]}
    scaneo_labeler.generate_stac_labels(
        catalog=tmp_stac_catalog,
        root_folder=os.path.join(EXAMPLE_DATA_DIR, "labels_scaneo"),
        collection="sentinel-2-l2a",
        **labels_extra_properties
    )

    assert tmp_stac_catalog.get_child("labels")
    labels_collection = tmp_stac_catalog.get_child("labels")
    assert LabelExtension.has_extension(labels_collection)
    summaries = labels_collection.summaries.to_dict()
    assert summaries["label:classes"]
    assert summaries["label:type"] == "vector"

    items = labels_collection.get_all_items()
    for i in items:
        props = i.properties
        assert props["label:methods"] == ["manual"]
        assert props["label:properties"] == ["label"]
        i.make_asset_hrefs_absolute()
        asset = i.to_dict()["assets"]["labels"]
        assert os.path.exists(asset["href"])



"""
Utils
"""

from sentinelhub import DataCollection, MosaickingOrder, MimeType
from .evalscripts import EvalScripts


class OUTPUT_FORMAT:
    TIFF = MimeType.TIFF
    JPG = MimeType.JPG
    PNG = MimeType.PNG


class SHParameters:
    """
    Sentinel Hub Parameters base class
    """

    MAX_CLOUD_COVERAGE: float = None
    FIELDS = None
    MOSAICKING_ORDER = MosaickingOrder.MOST_RECENT
    EVALSCRIPT = None
    OUTPUT_FORMAT = MimeType.TIFF

    def __init__(self):
        pass


class SHS2L2AParameters(SHParameters):
    """
    Sentinel-2-L2A parameters
    """

    DATA_COLLECTION = DataCollection.SENTINEL2_L2A
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.SENTINEL_2_L2A
    FIELDS = {
        "include": ["id", "properties.datetime", "properties.eo:cloud_cover"],
        "exclude": [],
    }
    FILTER = None
    RESOLUTION = 10
    BASE_URL = "https://services.sentinel-hub.com"
    CLOUD_COVERAGE = True


class SHS2L1CParameters(SHParameters):
    """
    Sentinel-2-L1C parameters
    """

    DATA_COLLECTION = DataCollection.SENTINEL2_L1C
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.SENTINEL_2_L1C
    FIELDS = {
        "include": ["id", "properties.datetime", "properties.eo:cloud_cover"],
        "exclude": [],
    }
    FILTER = None
    RESOLUTION = 10
    BASE_URL = "https://services.sentinel-hub.com"
    CLOUD_COVERAGE = True


class SHS1Parameters(SHParameters):
    """
    Sentinel-1 parameters
    """

    DATA_COLLECTION = DataCollection.SENTINEL1
    EVALSCRIPT = EvalScripts.SENTINEL_1
    MOSAICKING_ORDER = MosaickingOrder.MOST_RECENT
    FIELDS = {
        "include": [
            "id",
            "properties.datetime",
            "sar:instrument_mode",
            "s1:polarization",
            "sat:orbit_state",
            "s1:resolution",
            "s1:timeliness",
        ],
        "exclude": [],
    }
    FILTER = None
    RESOLUTION = 3
    BASE_URL = "https://services.sentinel-hub.com"
    CLOUD_COVERAGE = False


class SHDEMParameters(SHParameters):
    """
    Copernicus DEM parameters
    """

    DATA_COLLECTION = DataCollection.DEM_COPERNICUS_30
    MOSAICKING_ORDER = None
    EVALSCRIPT = EvalScripts.DEM
    FIELDS = None
    FILTER = None
    RESOLUTION = 3
    BASE_URL = "https://services.sentinel-hub.com"
    CLOUD_COVERAGE = False


class SHHarmonizedLandsatSentinel(SHParameters):
    """
    Harmonized Landsat Sentinel parameters
    """

    DATA_COLLECTION = DataCollection.HARMONIZED_LANDSAT_SENTINEL
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.HLS_TRUE_COLOR
    FIELDS = None
    FILTER = None
    RESOLUTION = 10
    BASE_URL = "https://services-uswest2.sentinel-hub.com"
    CLOUD_COVERAGE = True


class SHLandsatOTL2(SHParameters):
    """
    Landsat 8-9 Collection 2 imagery processed to level 2
    """

    DATA_COLLECTION = DataCollection.LANDSAT_OT_L2
    MOSAICKING_ORDER = MosaickingOrder.LEAST_CC
    EVALSCRIPT = EvalScripts.LANDSAT_OT_L2_TRUE_COLOR
    FIELDS = None
    FILTER = None
    RESOLUTION = 10
    BASE_URL = "https://services-uswest2.sentinel-hub.com"
    CLOUD_COVERAGE = True


class DATA_COLLECTION_ID:
    SENTINEL_1_GRD = DataCollection.SENTINEL1.api_id
    SENTINEL_2_L1C = DataCollection.SENTINEL2_L1C.api_id
    SENTINEL_2_L2A = DataCollection.SENTINEL2_L2A.api_id
    DEM = DataCollection.DEM_COPERNICUS_30.api_id
    HLS = DataCollection.HARMONIZED_LANDSAT_SENTINEL.api_id
    LANDSAT_OT_L2 = DataCollection.LANDSAT_OT_L2.api_id


SUPPORTED_COLLECTION_IDS = [
    value
    for name, value in DATA_COLLECTION_ID.__dict__.items()
    if not name.startswith("__")
]

SH_PARAMETERS_DICT = {
    DATA_COLLECTION_ID.SENTINEL_1_GRD: SHS1Parameters,
    DATA_COLLECTION_ID.SENTINEL_2_L1C: SHS2L1CParameters,
    DATA_COLLECTION_ID.SENTINEL_2_L2A: SHS2L2AParameters,
    DATA_COLLECTION_ID.DEM: SHDEMParameters,
    DATA_COLLECTION_ID.HLS: SHHarmonizedLandsatSentinel,
    DATA_COLLECTION_ID.LANDSAT_OT_L2: SHLandsatOTL2,
}


def get_default_parameters(collection_id: str) -> SHParameters:
    return SH_PARAMETERS_DICT[collection_id]()


def supports_cloud_coverage(collection_id: str):
    return SH_PARAMETERS_DICT[collection_id]().CLOUD_COVERAGE



from .AuthRepo import AuthRepo
from .APIRepo import APIRepo
from .AuthAPIRepo import AuthAPIRepo
from .DatasetsAPIRepo import DatasetsAPIRepo
from .FilesAPIRepo import FilesAPIRepo
from .ModelsAPIRepo import ModelsAPIRepo
from .STACAPIRepo import STACAPIRepo


import os

from datetime import datetime
from eotdl.curation import STACDataFrame
from eotdl.curation.stac.dataframe import read_stac

from .fixtures import tmp_stac_catalog, tmp_stac_collection


output_path = "tmp/output_path"


def test_from_stac_file(tmp_stac_catalog):
    df = STACDataFrame.from_stac_file(tmp_stac_catalog)
    assert isinstance(df, STACDataFrame)


def test_to_stac(tmp_stac_catalog):
    df = STACDataFrame.from_stac_file(tmp_stac_catalog)
    df.to_stac(output_path)
    assert os.path.exists(output_path)


def test_curate_json_row():
    sample_row = {
        "created_at": datetime.now(),
        "modified_at": datetime.now(),
        "stac_id": "123",
        "id": "456",
    }
    curated_row = STACDataFrame().curate_json_row(sample_row, True)
    # Assertions to ensure unnecessary keys are removed, "id" key exists, etc.
    assert "id" in curated_row
    assert "stac_id" not in curated_row


def test_read_stac(tmp_stac_catalog):
    df = read_stac(tmp_stac_catalog)
    assert isinstance(df, STACDataFrame)



import typer
import os

from .commands import auth, datasets, models, stac
from .repos import APIRepo
from . import __version__

app = typer.Typer(help="Welcome to EOTDL. Learn more at https://www.eotdl.com/")

app.add_typer(auth.app, name="auth")
app.add_typer(datasets.app, name="datasets")
app.add_typer(models.app, name="models")
app.add_typer(stac.app, name="stac")

@app.command()
def version():
    """
    Get EOTDL version.
    """
    typer.echo(f"EOTDL Version: {__version__}")


@app.command()
def api():
    """
    Get EOTDL API URL and info.
    """
    repo = APIRepo()
    typer.echo(f"EOTDL API URL: {repo.url}")
    typer.echo(repo.get_info())


if __name__ == "__main__":
    app()


from ..repos import ModelsAPIRepo


def update_model(model_id, metadata, content, user):
    repo = ModelsAPIRepo()
    data, error = repo.update_model(
        model_id,
        metadata.authors,
        metadata.source,
        metadata.license,
        metadata.thumbnail,
        content,
        user,
    )
    if error:
        raise Exception(error)
    return data

def deactivate_model(model_id):
    repo = ModelsAPIRepo()
    data, error = repo.deactivate_model(model_id)
    if error:
        raise Exception(error)
    return data



from glob import glob
from pathlib import Path
import geopandas as gpd
import os
import pystac
from tqdm import tqdm
import stac_geoparquet
import frontmatter
import random
import pandas as pd
from datetime import datetime
from shapely.geometry import Polygon

from ..auth import with_auth
from ..files.metadata import Metadata
from ..repos import FilesAPIRepo
from ..shared import calculate_checksum

def prep_ingest_folder(
	folder,
	verbose=False,
	logger=print,
	force_metadata_update=False,
	sync_metadata=False,
):
	logger("Ingesting directory: " + str(folder))
	catalog_path = folder.joinpath("catalog.parquet")
	files = glob(str(folder) + '/**/*', recursive=True)
	# remove catalog.parquet from files
	files = [f for f in files if f != str(catalog_path)]
	# ingest geometry from files (if tifs) or additional list of geometries
	# https://stac-utils.github.io/stac-geoparquet/latest/spec/stac-geoparquet-spec/#use-cases
	data = []
	for file in files:
		file_path = Path(file)
		if file_path.is_file():
			relative_path = os.path.relpath(file_path, catalog_path.parent)
			absolute_path = str(file_path)
			# THIS IS THE MINIMUM REQUIRED FIELDS TO CREATE A VALID STAC ITEM
			data.append(create_stac_item(relative_path, absolute_path))
	gdf = gpd.GeoDataFrame(data, geometry='geometry')
	# Save to parquet
	gdf.to_parquet(catalog_path)
	return catalog_path

# IF THE KEYS IN THE ASSETS ARE NOT THE SAME ON ALL ITEMS, THE PARQUET WILL NOT BE VALID !!!
def prep_ingest_stac(path, logger=None): # in theory should work with a remote catalog (given URL)
	# read stac catalog
	stac_catalog = path / "catalog.json"
	catalog = pystac.Catalog.from_file(stac_catalog)
	# make all items paths hredf in assets absolute
	catalog.make_all_asset_hrefs_absolute() 
	# generate list of items for all collections
	items = []
	for collection in catalog.get_collections():
		# iterate over items
		for item in tqdm(collection.get_items(), desc=f"Ingesting items from collection {collection.id}"):
			assert isinstance(item, pystac.Item)
			# Process each asset in the item
			for asset in item.assets.values():
				if not asset.href.startswith(('http://', 'https://')):
					# Asset is a local file
					file_path = Path(asset.href)
					# Calculate and add file size
					asset.extra_fields['size'] = file_path.stat().st_size
					# Calculate and add checksum
					asset.extra_fields['checksum'] = calculate_checksum(str(file_path))
			items.append(item)
	# save parquet file
	record_batch_reader = stac_geoparquet.arrow.parse_stac_items_to_arrow(items)
	output_path = stac_catalog.parent / "catalog.parquet"
	stac_geoparquet.arrow.to_parquet(record_batch_reader, output_path)
	return output_path

def ingest_virtual( # could work for a list of paths with minimal changes...
	path,
	links,
	repo,
	retrieve,
	mode,
	metadata = None, 
	logger=print,
):
	path = Path(path)
	if metadata is None:
		readme = frontmatter.load(path.joinpath("README.md"))
		metadata_dict = readme.metadata
		# Add description from content before creating Metadata object
		metadata_dict["description"] = readme.content
		metadata = Metadata(**metadata_dict)
	else:
		metadata = Metadata(**metadata)
		metadata.save_metadata(path)
	data = []
	for link in links:
		assert link.startswith("http"), "All links must start with http or https"
		data.append(create_stac_item(link, link))
	data.append(create_stac_item('README.md', str(path / "README.md")))
	gdf = gpd.GeoDataFrame(data, geometry='geometry')
	gdf.to_parquet(path / "catalog.parquet")
	return ingest(path, repo, retrieve, mode)

def ingest_catalog(path, repo, retrieve, mode):
	return ingest(path, repo, retrieve, mode)

@with_auth
def ingest(path, repo, retrieve, mode, user):
	try:
		readme = frontmatter.load(path.joinpath("README.md"))
		metadata_dict = readme.metadata
		# Add description from content before creating Metadata object
		metadata_dict["description"] = readme.content
		metadata = Metadata(**metadata_dict)
	except Exception as e:
		print(str(e))
		raise Exception("Error loading metadata")
	# retrieve dataset (create if doesn't exist)
	dataset_or_model = retrieve(metadata, user)
	current_version = sorted([v['version_id'] for v in dataset_or_model["versions"]])[-1]
	# TODO: update README if metadata changed in UI (db)
	# update_metadata = True
	# if "description" in dataset:
	#     # do not do this if the dataset is new, only if it already exists
	#     update_metadata = check_metadata(
	#         dataset, metadata, content, force_metadata_update, sync_metadata, folder
	#     )
	# if update_metadata:
	#     update_dataset(dataset["id"], metadata, content, user)
	# return ingest_files(
	#     repo, dataset["id"], folder, verbose, logger, user, endpoint="datasets"
	# )
	catalog_path = path.joinpath("catalog.parquet")
	gdf = gpd.read_parquet(catalog_path)
	files_repo = FilesAPIRepo()
	catalog_url = files_repo.generate_presigned_url(f'catalog.v{current_version}.parquet', dataset_or_model['id'], user, endpoint=mode)
	# first time ingesting
	if catalog_url is None:
		total_size = 0
		for row in tqdm(gdf.iterrows(), total=len(gdf), desc="Ingesting files"):
			try:
				for k, v in row[1]["assets"].items():
					if v["href"].startswith("http"): continue
					item_id = row[1]["id"]
					data, error = files_repo.ingest_file(
						v["href"],
						item_id, 
						# Path(v["href"]).stat().st_size,
						dataset_or_model['id'],
						user,
						mode,
					)
					if error:
						raise Exception(error)
					file_url = f"{repo.url}{mode}/{dataset_or_model['id']}/stage/{item_id}"
					gdf.loc[row[0], "assets"][k]["href"] = file_url
					total_size += v["size"]
			except Exception as e:
				print(f"Error uploading asset {row[0]}: {e}")
				break
		gdf.to_parquet(catalog_path)
		files_repo.ingest_file(str(catalog_path), f'catalog.v{current_version}.parquet', dataset_or_model['id'], user, mode)
		data, error = repo.complete_ingestion(dataset_or_model['id'], current_version, total_size, user)
		if error:
			raise Exception(error)
		return catalog_path
	
	# files were already ingested
	# TODO: check for deleted files (currently only updating existing files and ingesting new ones)
	# TODO: adding new links in virtual datasets dont trigger new version (but changing README does)
	new_version = False
	num_changes = 0
	total_size = 0
	for row in tqdm(gdf.iterrows(), total=len(gdf), desc="Ingesting files"):
		try:
			for k, v in row[1]["assets"].items():
				if v["href"].startswith("http"): continue
				item_id = row[1]["id"]
				# check if file exists in previous versions
				df = pd.read_parquet(
					path=catalog_url,
					filters=[('id', '=', item_id)]
				)
				if len(df) > 0: # file exists in previous versions
					if df.iloc[0]['assets'][k]["checksum"] == v["checksum"]: # file is the same
						# still need to update the required fields
						file_url = f"{repo.url}{mode}/{dataset_or_model['id']}/stage/{item_id}"
						gdf.loc[row[0], "assets"][k]["href"] = file_url
						total_size += v["size"]
						continue
					else: # file is different, so ingest new version but with a different id
						item_id = item_id + f"-{random.randint(1, 1000000)}"
						gdf.loc[row[0], "id"] = item_id
				new_version = True
				num_changes += 1
				# ingest new files
				data, error = files_repo.ingest_file(
					v["href"],
					item_id, # item id, will be path in local or given id in STAC. if not unique, will overwrite previous file in storage
					# Path(v["href"]).stat().st_size,
					dataset_or_model['id'],
					user,
					# calculate_checksum(asset["href"]),  # is always absolute?
					mode,
					# version,
				)
				if error:
					raise Exception(error)
				file_url = f"{repo.url}{mode}/{dataset_or_model['id']}/stage/{item_id}"
				gdf.loc[row[0], "assets"][k]["href"] = file_url
				total_size += v["size"]
		except Exception as e:
			print(f"Error uploading asset {row[0]}: {e}")
			break
	if not new_version:
		print("No new version was created, your dataset has not changed.")
	else:
		new_version = current_version + 1
		print("A new version was created, your dataset has changed.")
		print(f"Num changes: {num_changes}")
		gdf.to_parquet(catalog_path)
		files_repo.ingest_file(str(catalog_path), f'catalog.v{new_version}.parquet', dataset_or_model['id'], user, mode)
		# TODO: ingest README.md
		data, error = repo.complete_ingestion(dataset_or_model['id'], new_version, total_size, user)
		if error:
			raise Exception(error)
		return catalog_path

def create_stac_item(item_id, asset_href):
	return {
		'type': 'Feature',
		'stac_version': '1.0.0',
		'stac_extensions': [],
		'datetime': datetime.now(),  # must be native timestamp (https://github.com/apache/parquet-format/blob/master/LogicalTypes.md#timestamp)
		'id': item_id,
		'bbox': {
			'xmin': 0.0,
			'ymin': 0.0,
			'xmax': 0.0,
			'ymax': 0.0
		}, # infer from file or from list of geometries
		'geometry': Polygon(), # empty polygon
		'assets': { 'asset': { # STAC needs this to be a Dict[str, Asset], not list !!! use same key or parquet breaks !!!
			'href': asset_href,
			'checksum': calculate_checksum(asset_href) if not asset_href.startswith("http") else None,
			'timestamp': datetime.now(),
			'size': Path(asset_href).stat().st_size if not asset_href.startswith("http") else None,
		}},
		"links": [],
		# 'collection': 'source',
		# anything below are properties (need at least one!)
		'repository': 'eotdl',				
	}


from eotdl.tools import count_ocurrences, cut_images, get_all_images_in_path


# Test for count_ocurrences function
def test_count_ocurrences():
    test_list = ["apple", "orange", "applejuice", "applepie", "banana"]
    assert count_ocurrences("apple", test_list) == 3
    assert count_ocurrences("orange", test_list) == 1
    assert count_ocurrences("mango", test_list) == 0


# Test for cut_images function
def test_cut_images():
    test_images_list = [
        "/path/to/dir1/image1.tif",
        "/path/to/dir1/image2.tif",
        "/path/to/dir2/image1.tif",
        "/path/to/dir3/image1.tif"
    ]
    cut_list = cut_images(test_images_list)
    assert len(cut_list) == 3
    assert "/path/to/dir1/image1.tif" in cut_list
    assert "/path/to/dir2/image1.tif" in cut_list
    assert "/path/to/dir3/image1.tif" in cut_list


# Test for get_all_images_in_path using pytest's tmpdir fixture
def test_get_all_images_in_path(tmpdir):
    # Create dummy image files in tmpdir
    sub_dir = tmpdir.mkdir("sub")
    sub_dir.join("image1.tif").write("content")
    sub_dir.join("image2.tif").write("content")
    tmpdir.join("image3.tif").write("content")

    # Get all tif images in tmpdir
    images = get_all_images_in_path(str(tmpdir))
    assert len(images) == 3
    assert any("image1.tif" in path for path in images)
    assert any("image2.tif" in path for path in images)
    assert any("image3.tif" in path for path in images)



from pathlib import Path

from ..repos import DatasetsAPIRepo
from ..files.ingest import prep_ingest_stac, prep_ingest_folder, ingest, ingest_virtual, ingest_catalog

def retrieve_dataset(metadata, user):
	repo = DatasetsAPIRepo()
	data, error = repo.retrieve_dataset(metadata.name)
	if data and data["uid"] != user["uid"]:
		raise Exception("Dataset already exists.")
	if error and error == "Dataset doesn't exist":
		# create dataset
		data, error = repo.create_dataset(metadata.dict(), user)
		# print(data, error)
		if error:
			raise Exception(error)
	return data

def ingest_dataset(
	path,
	verbose=False,
	logger=print,
	force_metadata_update=False,
	sync_metadata=False,
):
	path = Path(path)
	if not path.is_dir():
		raise Exception("Path must be a folder")
	if "catalog.json" in [f.name for f in path.iterdir()]:
		prep_ingest_stac(path, logger)
	else:
		prep_ingest_folder(path, verbose, logger, force_metadata_update, sync_metadata)
	return ingest(path, DatasetsAPIRepo(), retrieve_dataset, 'datasets')


def ingest_virtual_dataset( # could work for a list of paths with minimal changes...
	path,
	links,
	metadata = None, 
	logger=print,
	user=None,
):
	return ingest_virtual(path, links, DatasetsAPIRepo(), retrieve_dataset, 'datasets', metadata, logger)

def ingest_dataset_catalog(path, logger=None):
	path = Path(path)
	return ingest_catalog(path, DatasetsAPIRepo(), retrieve_dataset, 'datasets')


from ..repos import DatasetsAPIRepo


def update_dataset(dataset_id, metadata, content, user):
    repo = DatasetsAPIRepo()
    data, error = repo.update_dataset(
        dataset_id,
        metadata.authors,
        metadata.source,
        metadata.license,
        metadata.thumbnail,
        content,
        user,
    )
    if error:
        raise Exception(error)
    return data


def deactivate_dataset(dataset_id):
    repo = DatasetsAPIRepo()
    data, error = repo.deactivate_dataset(dataset_id)
    if error:
        raise Exception(error)
    return data


import time
import os

from ..repos import AuthRepo, AuthAPIRepo
from .errors import LoginError, AuthTimeOut
from .is_logged import is_logged


def auth(max_t=60, interval=2):
    user = is_logged()
    if user:
        return user
    repo, api_repo = AuthRepo(), AuthAPIRepo()
    api_key = os.environ.get("EOTDL_API_KEY", None)
    if api_key:
        print("Using API Key")
        user = {"api_key": api_key}
        user_data, error = api_repo.retrieve_user_data(user)
        if error:
            raise LoginError()
        user.update({"email": user_data["email"], "uid": user_data["uid"]})
    else:
        response = api_repo.login()
        if response.status_code != 200:
            raise LoginError()
        data = response.json()
        print("On your computer or mobile device navigate to: ", data["login_url"])
        authenticated = False
        t0 = time.time()
        while not authenticated and time.time() - t0 < max_t:
            response = api_repo.token(data["code"])
            token_data = response.json()
            if response.status_code == 200:
                print("Authenticated!")
                print("- Id Token: {}...".format(token_data["id_token"][:10]))
                user = repo.decode_token(token_data)
                authenticated = True
                user = {
                    "id_token": token_data["id_token"],
                    "email": user["email"],
                    "uid": user["sub"],
                }
            else:
                time.sleep(interval)
        if not authenticated:
            raise AuthTimeOut()
    # get user credentials
    credentials = api_repo.retrieve_credentials(user)[0]
    if credentials:
        user.update(credentials)
    # save token data in file
    creds_path = repo.save_creds(user)
    print("Saved credentials to: ", creds_path)
    return user


# auth decorator
def with_auth(func):
    def wrapper(*args, **kwargs):
        user = auth()
        return func(*args, **kwargs, user=user)

    return wrapper



"""
Module for managing the Airbus configuration and data access
"""

import json
import time
from typing import Optional, Iterable
from os.path import join, exists
import requests
from requests.exceptions import ConnectTimeout, ReadTimeout

from .parameters import AirbusURL, AirbusProductType, AirbusImageFormat, AirbusRadiometricProcessing
from .utils import get_airbus_access_token
from ...tools import expand_time_interval, bbox_to_coordinates


class AirbusClient:
    """
    Client class to manage the Sentinel Hub Python interface.
    """

    def __init__(
        self,
        access_token: str,
        api_key: str,
    ) -> None:
        """
        Constructor

        Params
        ----------
        access_token: str
            Access token
        api_key: str
            API key
        """
        self.airbus_access_token = access_token
        self._api_key = api_key

    def get_total_products_price(
        self, payload: dict, all_info: Optional[bool] = False
    ) -> dict:
        """
        Get total products price
        """
        if all_info:
            response = []
        else:
            response = 0

        for _, location_data in payload.items():
            product_id = location_data["image"]
            if product_id:
                price_response = self.get_product_price(
                    product_id, location_data["bounding_box"]
                )
                if all_info:
                    response.append(price_response)
                else:
                    response += price_response["price"]["credits"]

        return response

    def get_product_price(
        self,
        product_id: str,
        coordinates: Iterable,
        product_type: Optional[AirbusProductType] = AirbusProductType.MULTISPECTRAL,
        image_format: Optional[AirbusImageFormat] = AirbusImageFormat.GEOTIFF,
        processing: Optional[
            AirbusRadiometricProcessing
        ] = AirbusRadiometricProcessing.REFLECTANCE,
    ) -> dict:
        """
        Get product price

        Params
        ----------
        product_id: str
            Product ID
        coordinates: tuple
            Polygon coordinates
        product_type: AirbusProductType
            Product type
        image_format: AirbusImageFormat
            Image format
        processing: AirbusRadiometricProcessing
            Radiometric processing
        Types are defined at parameters.py

        Returns
        ----------
        dict
            Product price
        """
        if isinstance(coordinates, (list, tuple)) and len(
            coordinates
        ) == 4:
            coordinates = bbox_to_coordinates(coordinates)

        headers = {
            "Authorization": f"Bearer {self.airbus_access_token}",
            "Content-Type": "application/json",
            "Cache-Control": "no-cache",
        }

        # TODO make productType, imageformat and processing configurable with paramters controlled by classes like
        # TODO AirbusProductType.multiSpectral, AirbusImageFormat.geotiff, AirbusProcessing
        payload = {
            "kind": "order.data.product",
            "products": [
                {
                    "productType": product_type,
                    "radiometricProcessing": processing,
                    "imageFormat": image_format,
                    "crsCode": "urn:ogc:def:crs:EPSG::4326",
                    "id": product_id,
                    "aoi": {"type": "Polygon", "coordinates": [coordinates]},
                }
            ],
        }
        response = requests.request(
            "POST", AirbusURL.PRICES, json=payload, headers=headers, timeout=60000
        )

        return response.json()

    def place_product_order(
        self,
        product_id: str,
        coordinates: Iterable,
        product_type: Optional[AirbusProductType] = AirbusProductType.MULTISPECTRAL,
        image_format: Optional[AirbusImageFormat] = AirbusImageFormat.GEOTIFF,
        processing: Optional[
            AirbusRadiometricProcessing
        ] = AirbusRadiometricProcessing.REFLECTANCE,
    ) -> dict:
        """
        Place product order

        Params
        ----------
        product_id: str
            Product ID
        bounding_box: tuple
            Bounding box
        product_type: AirbusProductType
            Product type
        image_format: AirbusImageFormat
            Image format
        processing: AirbusRadiometricProcessing
            Radiometric processing
        Types are defined at parameters.py

        Returns
        ----------
        dict
            Order data
        """
        if isinstance(coordinates, (list, tuple)) and len(
            coordinates
        ) == 4:
            coordinates = bbox_to_coordinates(coordinates)

        payload = {
            "kind": "order.data.product",
            "products": [
                {
                    "productType": product_type,
                    "radiometricProcessing": processing,
                    "imageFormat": image_format,
                    "crsCode": "urn:ogc:def:crs:EPSG::4326",
                    "id": product_id,
                    "aoi": {"type": "Polygon", "coordinates": [coordinates]},
                }
            ],
        }

        headers = {
            "Authorization": f"Bearer {self.airbus_access_token}",
        }

        response = requests.request(
            "POST", AirbusURL.ORDERS, json=payload, headers=headers, timeout=60000
        )

        return response.json()

    def search_image(
        self,
        bounding_box: Iterable,
        acquisition_date: Iterable,
        timeout: Optional[int] = 10,
    ) -> dict:
        """
        Search image

        Params
        ----------
        bounding_box: tuple or list
            Bounding box
        acquisition_date: tuple or list
            Acquisition date
        timeout: int
            Timeout

        Returns
        ----------
        dict
            Image data
        """
        if isinstance(acquisition_date, (tuple, list)):
            acquisition_date = "[" + ",".join(acquisition_date) + "]"
        if isinstance(bounding_box, (tuple, list)) and len(
            bounding_box
        ) == 4:
            bounding_box = ",".join(str(num) for num in bounding_box)

        querystring = {"acquisitionDate": str(acquisition_date), "bbox": bounding_box}

        headers = {
            "authorization": f"Bearer {self.airbus_access_token}",
            "cache-control": "no-cache",
        }

        try:
            response = requests.request(
                "GET",
                AirbusURL.SEARCH,
                headers=headers,
                params=querystring,
                verify=False,
                timeout=timeout,
            )
            return response.json()
        except json.decoder.JSONDecodeError:
            print("JSONDecodeError")
            print(response)
        except ReadTimeout:
            print("ReadTimeout")
            print(response)

    def search_images_close_in_time(
        self,
        payload_dict: dict,
        path: Optional[str] = None,
        max_days: Optional[int] = 30,
    ) -> dict:
        """
        Search images close in time

        Params
        ----------
        payload_dict: dict
            Payload dictionary
        max_days: int
            Maximum days to search

        Returns
        ----------
        dict
            Dictionary with the image data, as {location_id: image_data}, to
            maintain track of the location
        """
        responses = {}
        if path:
            responses_path = join(path, "airbus_images_response.json")
            if exists(responses_path):
                with open(responses_path, "r", encoding="utf-8") as f:
                    responses = json.load(f)

        for location_id, location_info in list(payload_dict.items()):
            bounding_box, time_interval = (
                location_info["bounding_box"],
                location_info["time_interval"],
            )
            days = 1

            if location_id in responses:
                continue

            while days <= max_days:
                try:
                    response = self.search_image(bounding_box, time_interval)
                except ConnectTimeout or TimeoutError:
                    # Wait 5 seconds and try again
                    time.sleep(5)
                    # Restart the connection
                    self.airbus_access_token = get_airbus_access_token(self._api_key)
                    # Continue with the search
                    response = self.search_image(bounding_box, time_interval)

                total_results = response["totalResults"]

                if total_results > 0:
                    # By default, the search results are sorted per acquisition date
                    # (newest data is displayed first) and per cloud coverage (less cloudy images are displayed first).
                    # So, we can stop the search when we find the first image
                    if total_results > 1:
                        response["features"] = [response["features"][0]]
                    responses[location_id] = response["features"][0]
                    if path:
                        with open(responses_path, "w", encoding="utf-8") as f:
                            json.dump(responses, f)
                    break

                time_interval = expand_time_interval(time_interval)
                days += 1

            # If no image is found, we add a None value to the dictionary
            if total_results == 0:
                responses[location_id] = None
                if path:
                    with open(responses_path, "w", encoding="utf-8") as f:
                        json.dump(responses, f)

        return responses

    def format_product_payload(
        self, location_payload: dict, images_response: dict
    ) -> dict:
        """
        Format product payload
        """
        for product_id, _ in location_payload.items():
            # Add new key to the dictionary
            location_payload[product_id]["image_response"] = (
                images_response[product_id] if product_id in images_response else None
            )

        return location_payload

    def split_product_payload(self, product_payload: dict) -> dict:
        """
        Split product payload
        """
        # split the product payload depending on if 'image' is None or not
        product_payload_with_image = {}
        product_payload_without_image = {}

        for product_id, info in product_payload.items():
            if info["image"]:
                product_payload_with_image[product_id] = info
            else:
                product_payload_without_image[product_id] = info

        return product_payload_with_image, product_payload_without_image

    def get_all_order_status(self):
        """
        Get all order status
        """
        headers = {
            "Authorization": f"Bearer {self.airbus_access_token}",
            "Content-Type": "application/json",
            "Cache-Control": "no-cache",
        }

        response = requests.request("GET", AirbusURL.ALL_ORDERS_STATUS, headers=headers, timeout=60000)

        return response.text

    def get_account_information(self):
        """
        Get account information
        """
        headers = {
            "Authorization": f"Bearer {self.airbus_access_token}",
            "Content-Type": "application/json",
            "Cache-Control": "no-cache",
        }

        response = requests.request("GET", AirbusURL.ACCOUNT, headers=headers, timeout=60000)

        return response.json()

    def get_user_roles(self):
        """
        Get user roles
        """
        headers = {"Authorization": f"Bearer {self.airbus_access_token}"}

        response = requests.request("GET", AirbusURL.ROLES, headers=headers, timeout=60000)

        return response.json()
